import math, cmath, scipy.integrate
import numpy as np

# Transformation conforme
def inversion_domaine(centres, rayons, delta):
    c_t = [(z-delta).conjugate()/(abs(z-delta)*abs(z-delta) - r*r) for z, r in zip(centres, rayons)]
    r_t = [r/abs(abs(z-delta)*abs(z-delta) - r*r) for z, r in zip(centres, rayons)]
    return c_t, r_t


# Appartenance d'un point au domaine
eps = 1.e-9

def point_exterieur_T(z, centres, rayons):
    for c, r in zip(centres, rayons):
        if abs(z-c)<r:
            return True
    return False

def point_exterieur(z, centres, rayons, eps=eps):
    """Si vrai, alors le point n'appartient pas au domaine.

    Cas du domaine de d\'epart o\`u le premier cercle de la liste est
    ext\'erieur.
    
    Pr\'evoit une marge pour que les points du bord soient marqu\'es faux.
    """
    if abs(z-centres[0])>rayons[0]+eps:
        return True
    for c, r in zip(centres[1:], rayons[1:]):
        if abs(z-c)<r-eps:
            return True
    return False

def point_interieur(z, centres, rayons, eps=eps):
    """Si vrai, alors le point appartient au domaine.

    Cas du domaine de d\'epart o\`u le premier cercle de la liste est
    ext\'erieur.
    
    Pr\'evoit une marge pour que les points du bord soient marqu\'es faux.
    """
    if abs(z-centres[0])>rayons[0]-eps:
        return False
    for c, r in zip(centres[1:], rayons[1:]):
        if abs(z-c)<r+eps:
            return False
    return True

def point_bord(z, centres, rayons, eps=eps):
    """Vrai si le point appartient au domaine.

    Cas du domaine de d\'epart o\`u le premier cercle de la liste est
    ext\'erieur.
    
    Pr\'evoit une marge pour que les points du bord soient marqu\'es vrai.
    """
    if abs(abs(z-centres[0]) - rayons[0])<eps:
        return True
    for c, r in zip(centres[1:], rayons[1:]):
        if abs(abs(z-c)-r)<eps:
            return True
    return False
    
# Points de collocation
def points_collocation(centres, rayons, npts):
    """Placer des points sur les bords de cercles

    Parameters:
    -----------
    centres: list of cfloats
        valeur dans le plan complexe des centres
    rayons: list of floats
        rayons des cercles
    npts: int
        nombre de points de collocation par cercle

    Returns:
    --------
    z: list of cfloats
        points de collocation dans le plan complexe
    """
    racines_unite = [cmath.exp(k*2j*math.pi/npts) for k in np.arange(npts)]
    z = [c+r*u for c,r in zip(centres, rayons) for u in racines_unite]
    return z

 
# Moindres carres
def _fill_A(z, J, N, centres):
    res = np.ones((len(z), 1+J+2*J*N))
    l=1
    for cen in centres:
        for i, zi in enumerate(z):
            res[i, l] = math.log(abs(zi-cen))
        l+=1
    for cen in centres:
        for j in range(N):
            for i, zi in enumerate(z):
                tmp = (zi-cen)**(-j-1)
                res[i, l] = tmp.real
                res[i, l+1] = tmp.imag
            l+=2
    return res


def coeffs_harmonique(centres, N, z):
    """Calcul des  coefficients de l'approximation de la fonction harmonique

    Domaine non born\'e : que des cercle int\'erieurs

    Parameters:
    -----------
    
    centres: list of cfloats
        valeur dans le plan complexe des centres
    cl: list of floats
        valeur de la solution sur le cercle
    N: int
        nombre de termes dans le developpement
    z: list of cfloats
        points de collocation

    Returns:
    --------
    a: list of lists of floats
        partie reelle du DL
    b: list of lists of floats
        partie imaginaire du DL
    d: list of floats
        termes en log
    e: float
        terme constant
    """
    J = len(centres)
    
    cl = np.zeros(len(z))
    A = _fill_A(z, J, N, centres)
    
    rhs = np.zeros(2+J+2*N*J)
    for j, a in enumerate(A.transpose()): # par ligne de A^t
        rhs[j] = 2*np.dot(a.transpose(), cl)
    rhs[-1] = 1
    
    v = np.zeros(1+J+2*N*J)
    v[1:J+1] = np.ones(J) # \sum d_j = 1
    
    B = np.zeros((2+J+2*J*N,2+J+2*J*N))
    B[:-1,:-1] = np.dot(A.transpose(), 2*A)
    B[-1:, :-1] = v
    B[:-1, -1:] = v.reshape(1+J+2*N*J, 1)
    
    c = np.linalg.solve(B, rhs)
    resa = (c[J+1:-1:2]).reshape(J, N) 
    resb = (c[J+2::2]).reshape(J, N) 
    return resa, resb, c[1:J+1], c[0]

    
# Fonction solution harmonique
def harmonique_force_ext(centres, rayons, a, b, d, e):
    """Approximation de la fonction harmonique

    La valeur a l'interieur des cercles est fixee a cl.

    Parameters:
    -----------
    centres: list of cfloats
        valeur dans le plan complexe des centres
    rayons: list of floats
        rayons des cercles
    a: list of lists of floats
        partie reelle du DL
    b: list of lists of floats
        partie imaginaire du DL
    d: list of floats
        termes en log
    e: float
        terme constant

    Returns:
    --------
    g: fonction d'approximation
    """
    def _g(Z):
        if point_exterieur_T(Z, centres, rayons): return 0
        res = e
        for cen, di in zip(centres, d):
            res += di*math.log(abs(Z-cen))
        for cen, ai, bi in zip(centres, a, b):
            for j, (aij, bij) in enumerate(zip(ai, bi)):
                tmp = (Z-cen)**(-j-1)
                res += ((aij - 1j*bij)*tmp).real
        return res
    return _g

    
def harmonique(centres, a, b, d, e):
    """Approximation de la fonction harmonique

    La valeur sur les centres est fixee \`a 0 pour eviter les singularites
    """
    def _g(Z):
        if Z in centres: return 0
        res = e
        for cen, di in zip(centres, d):
            res += di*math.log(abs(Z-cen))
        for cen, ai, bi in zip(centres, a, b):
            for j, (aij, bij) in enumerate(zip(ai, bi)):
                tmp = (Z-cen)**(-j-1)
                res += ((aij - 1j*bij)*tmp).real
        return res
    return _g


def gradient_harmonique(centres, a,b,d):
    """Gradient de la fonction d\'efinie dans harmonique.
    """
    def _g(Z):
        x = Z.real
        y = Z.imag
        if Z in centres: return 0
        gx = 0
        gy = 0
        for cen, di in zip(centres, d):
            x0 = cen.real
            y0 = cen.imag
            den = (x-x0)*(x-x0) + (y-y0)*(y-y0)
            gx += di*(x-x0)/den
            gy += di*(y-y0)/den
        for cen, ai, bi in zip(centres, a, b):
            for j, (aij, bij) in enumerate(zip(ai, bi)):
                tmp = (aij+1j*bij)*((Z-cen)**(-j-1))
                gx += -j*(tmp).real
                gy += j*(tmp).imag
        return gx, gy
    return _g
    

# Fonction solution de Green
def fonction_green(centres, rayons, delta, centres_T, a,b,d,e):
    """Fonction de Green sur le domaine initial.

    G(delta, .)

    On impose G(delta, delta) = 0
    """
    def g(z):
        try:
            zt = 1./(z-delta)
        except ZeroDivisionError:
            return 0
        if point_exterieur(z, centres, rayons):
            raise ArithmeticError('Trying to compute outside of domain')
        harm = harmonique(centres_T, a,b,d,e)
        res = harm(zt)
        return res
    return g


def gradient_green(centres, delta, centres_T, a,b,d):
    """Gradient de la fonction de green

    \nabla G(delta, .)
    """
    grad_h = gradient_harmonique(centres_T, a,b,d)
    
    def g(Z):
        x = Z.real
        y = Z.imag
        
        try:
            zt = 1./(Z - delta)
        except ZeroDivisionError:
            hdx, hdy = 0, 0
        else:
            hdx, hdy = grad_h(zt)

        x0 = delta.real
        y0 = delta.imag
        den = ((x-x0)*(x-x0)+(y-y0)*(y-y0) )*( (x-x0)*(x-x0)+(y-y0)*(y-y0))
        txx = (-(x-x0)*(x-x0) + (y-y0)*(y-y0))/den # composantes de (Jac T)
        txy = (2*(x-x0)*(y-y0))/den
        
        resx = hdx*txx + hdy*txy
        resy =-hdx*txy + hdy*txx
        return resx, resy
        
    return g


def delta_green(centre, rayon, delta_rho, gradg_zeta, gradg_eta):
    """
    Variation de Hadamard de la fonction de Green sur \Omega \times \Omega

    Fonction \'evalu\'ee au point (zeta,eta)
        \delta G (\zeta, \eta)
    Cas o\`u un cercle interieur est modifi\'e.

    Parameters:
    -----------
    centre: cfloat
        centre du cercle modifi\'e
    rayon: float
        rayon du cercle modifi\'e
    delta_rho: [0, 2*pi] -> float
        perturbation du cercle
    gradg_zeta: cfloat -> float
        grad G(zeta, .)
    gradg_eta: cfloat -> float
        grad G(eta, .)

    Returns:
    --------
    dG(zeta, eta)
    """
    def integrand(t):
        zt = centre + rayon*cmath.exp(t*1j)
        ntx = -math.cos(t)
        nty = -math.sin(t)
        
        dgx_eta, dgy_eta = gradg_eta(zt)
        dgn_eta = dgx_eta*ntx + dgy_eta*nty
        dgx_zeta, dgy_zeta = gradg_zeta(zt)
        dgn_zeta = dgx_zeta*ntx + dgy_zeta*nty

        return rayon * delta_rho(t) * dgn_eta * dgn_zeta #* 0.5/math.pi

    return scipy.integrate.quad(integrand, 0, 2*math.pi)[0]


def delta_green_n(centre, rayon, delta_rho, g_zeta, g_eta, pas=1e-6):
    """
    Variation de Hadamard de la fonction de Green sur \Omega \times \Omega

    Ici, la derivee normale est approchee numeriquement
    """
    def integrand(t):
        nt = cmath.exp(t*1j)
        zt = centre + rayon*nt
        
        dgn_eta = (g_eta(zt+pas*nt) - g_eta(zt))/pas
        dgn_zeta =(g_zeta(zt+pas*nt) - g_zeta(zt))/pas

        return -rayon * delta_rho(t) * dgn_eta * dgn_zeta
        # le coef -1 pour parcourir le bord dans le sens horaire

    return scipy.integrate.quad(integrand, 0, 2*math.pi)[0]/2/math.pi
    
def green(centres, rayons, delta, npts, N):
    if npts < 2*N:
        raise ValueError('Not enough collocation points')
    centres_T, rayons_T = inversion_domaine(centres, rayons, delta)
    z_T = points_collocation(centres_T, rayons_T, npts)
    a,b,d,e = coeffs_harmonique(centres_T, N, z_T)
    g = fonction_green(centres, rayons, delta, centres_T, a,b,d,e)
    gradg = gradient_green(centres, delta, centres_T, a,b,d)
    return g, gradg
    
def eval_u(x0, u0, centres, rayons, npts, N):
    """ Valeur en $x0$ de la solution harmonique ayant pour CL $u0$

    Parameters:
    -----------
    x0: cfloat
    u0: float, int -> float
    centres: list(cfloat)
    rayons: list(float)
    npts: int
    N: int
    """
    g_sol, grad_g = green(centres, rayons, x0, npts, N)
    res = 0
    for j, (cen, ray) in enumerate(zip(centres, rayons)):
        def integrand(t):
            zt = cen + ray*cmath.exp(t*1j)
            ntx = math.cos(t)
            nty = math.sin(t)
            dgx, dgy = grad_g(zt)
            return u0(t, j) * (dgx*ntx + dgy*nty)
        tmp = ray * scipy.integrate.quad(integrand, 0, 2*math.pi)[0]
        if j==0:
            res -= tmp
        else:
            res += tmp
    return res*0.5/math.pi
                
def eval_grad_u(x0, u0, centres, rayons, npts, N, h):
    """ Valeur en $x0$ du gradient de la solution harmonique cf `eval_u`

    Attention si x0 est un point du bord !!!

    Parameters:
    -----------
    h: pas avec lequel on calcule le gradient num\'erique
    
    """
    zi = []
    for z1 in [x0+h, x0-h, x0+1j*h, x0-1j*h]:
        if point_exterieur(z1, centres, rayons):
            zi.append(x0)
        else:
            zi.append(z1)

    uz = []
    for z1 in zi:
        uz.append(eval_u(z1, u0, centres, rayons, npts, N))

    du_dx = (uz[0]-uz[1]) / abs(zi[0]-zi[1])
    du_dy = (uz[2]-uz[3]) / abs(zi[2]-zi[3])

    return du_dx, du_dy

def eval_delta_u(x0, centre, rayon, delta_rho, gradient_u, gradient_g):
    """ delta_u(z) = \int dr dg/dn du/dn ds

    Integrale sur un cercle

    Parameters:
    -----------
    x0: cfloat
        point en lequel on evalue delta_u
    centre: cfloat
        centre du cercle sur lequel on integre
    rayon: float
        rayon du cercle sur lequel on integre
    delta_rho: float -> float
        perturbation en coordonnees cylindriques le long du cercle (centre, rayon)
    gradient_u: cfloat -> (float, float)
        gradient de la solution harmonique u
    gradient_g: cfloat -> (float, float)
        gradient de la fonction de Green en x0
        nabla(g(x0, .))
    """
    def integrand(theta):
        zt = centre + rayon*cmath.exp(theta*1j)
        ntx = -math.cos(theta)
        nty = -math.sin(theta)
        
        udx, udy = gradient_u(zt)
        du_dn = udx*ntx + udy*nty
        
        dgx, dgy = gradient_g(zt)
        dg_dn = dgx*ntx + dgy*nty
        
        return delta_rho(theta) * du_dn * dg_dn
        
    res = scipy.integrate.quad(integrand, 0, 2*math.pi, limit=20)[0]
    # Le /4/math.pi est mathemagique
    return res/4./math.pi
